// $Id: StopTimer.h,v 1.3 2009/05/25 11:58:40 mwache Exp $
#ifndef CORALSERVERBASE_STOPTIMER_H 
#define CORALSERVERBASE_STOPTIMER_H 1

// Include files
#include <string>
#include <sys/times.h>


namespace coral {
  /**
    * Usage:
    *
    * StopTimer myTimer;
    * myTimer.start();
    * // performance critical part
    * myTimer.stop("package::class::performancePart1")
    * 
    * The times will be automatically added to the report
    *
    * Alternatively you can use the scoped timer, which is very nice for
    * functions:
    * 
    * void class::doSomething()
    * {
    *   ScopedTimer timer("package::class::doSomething");
    *   // do something
    * }
    *
    *
    */

  /// globaly enable/disable timings
  /// by default the timings are disabled, and can be enabled
  /// by setting the environment variable CORALSERVER_TIMING
  void enableTimings( bool enabled=true );

  /// prints times of all registered timers
  void printTimers();

  /// deletes all so far collected statistics
  void deleteTimings();

  class StopTimer {
    public:
      /// default constructor
      StopTimer();

      virtual ~StopTimer() {};

      /// start timer 
      virtual void start( /*const std::string& name*/);

      /// stop timer and add to report
      virtual void stop( const std::string& name );

      /// get user time in seconds
      virtual double getUserTime() const;
        
      /// get system time in seconds
      virtual double getSystemTime() const;

      /// get the real time in seconds
      virtual double getRealTime() const;

    private:
      /// is the watch running?
      bool m_isRunning;

      /// how much time elapsed
      struct tms m_elapsedTime;
   
      /// starting time of the watch
      struct tms m_startTime;

      /// elapsed real time
      clock_t m_elapsedReal;

      /// starting real time
      clock_t m_startReal;

#ifdef __linux
      unsigned long m_startUserLinux;
      unsigned long m_startSystemLinux;
      unsigned long m_elapsedUserLinux;
      unsigned long m_elapsedSystemLinux;
#endif
  };

  class ScopedTimer {
    public:
      ScopedTimer( const std::string& name )
        : m_name( name ), m_stopTimer()
      {
        m_stopTimer.start();
      };

      ~ScopedTimer()
      {
        m_stopTimer.stop( m_name );
      };
    private:
      std::string m_name;
      StopTimer m_stopTimer;
  };

}

#endif
