// $Id: SocketRequestHandler.cpp,v 1.10 2009/07/10 09:04:20 avalassi Exp $

// Include files
#include <iostream>

#include "SocketRequestHandler.h"

// Local include files
#include "PacketSocket.h"
#include "ReplyManager.h"

#define LOGGER_NAME "CoralSockets::SocketRequestHandler"
#include "logger.h"

// Namespace
using namespace coral::CoralSockets;
using namespace coral;

//-----------------------------------------------------------------------------

namespace coral {

  namespace CoralSockets {


    /// a class which does the receiving of packets in an own thread
    class SocketRequestHandler::receiveThread {
      public:
        receiveThread( const PacketSocketPtr& socket, ReplyManager& rplMngr ) :
          m_Socket( socket ),
          m_rplMngr( rplMngr)
        {
          DEBUG("socketRequestHandler::receiveThread constructor");
        };

        virtual ~receiveThread()
        {
          DEBUG("socketRequestHandler::receiveThread destructor");
        };

        /// the main loop of the thread
        void operator()();

      private:
        /// the socket which is used to receive data
        const PacketSocketPtr& m_Socket;

        /// storage container for the waiting requests
        ReplyManager& m_rplMngr;
    };

  }
} // namespace CORAL & CoralSockets

//-----------------------------------------------------------------------------

// the main loop of the receiver thread
void SocketRequestHandler::receiveThread::operator()() {
  INFO("SocketRequestHandler::receiveThread main loop started");

  while ( true ) 
  { 
    try
    {
      PacketPtr packet( m_Socket->receivePacket() );

      const int requestID = packet->getHeader().requestID();
      DEBUG("Got reply to request #" << requestID );
      static int csdebug = -1; //AV
      if ( csdebug == -1 ) csdebug = ( getenv ( "CORALSOCKETS_DEBUG" ) ? 1 : 0 ); //AV
      if ( csdebug > 0 ) std::cout << "Got reply to request #" << requestID << std::endl; //AV

      // find sleeper slot
      ReplySlot &slot = m_rplMngr.findSlot( requestID );

      // put reply and wake up the sleeper!
      slot.appendReply( packet->getPayloadPointer(), 
          packet->getHeader().segmentNumber(), 
          !packet->getHeader().moreSegments() /* isLast */ );

    }
    catch( SocketClosedException& e ) {
      INFO("The socket has been closed. Stopping threads.");
      break;
    }
    catch( std::exception& e ) {
      ERROR("ERROR! Standard C++ exception: '" << e.what() << "'");
      break;
    }
    catch( ... ) {
      ERROR("ERROR! Unknown exception caught");
      break;
    }
  }

  // notify waiting threads that the socket has been closed
  m_rplMngr.close();

  INFO("SocketRequestHandler::receiveThread main loop ended");
}

//-----------------------------------------------------------------------------

void 
SocketRequestHandler::setCertificateData( const coral::ICertificateData* /*cert*/ )
{
  // DUMMY ???
}

//-----------------------------------------------------------------------------

IByteBufferIteratorPtr 
SocketRequestHandler::replyToRequest( const ByteBuffer& requestBuffer )
{
  ReplySlot &slot = send( requestBuffer );
  return IByteBufferIteratorPtr( new SocketReplyIterator( slot ) );
}      

//-----------------------------------------------------------------------------

SocketRequestHandler::SocketRequestHandler( PacketSocketPtr socket )
  : IRequestHandler()
  , m_Socket( socket )
  , m_requestID( 0 )
  , m_rplMngr( new ReplyManager() )
{
  INFO("SocketRequestHandler constructor connection:"<< m_Socket->desc());

  receiveThread receiver( m_Socket, *m_rplMngr );
  m_receiveThread = std::auto_ptr<boost::thread>(new boost::thread( receiver ));
}

//-----------------------------------------------------------------------------

SocketRequestHandler::~SocketRequestHandler()
{
  INFO("Destructor SocketRequestHandler started");
  // FIXME is this racy? the receive thread accesses m_Socket!!!
  m_Socket->close();
  // wait until the receive thread has terminated
  m_receiveThread->join();
  DEBUG("Destructor SocketRequestHandler finished");
}

//-----------------------------------------------------------------------------

ReplySlot& SocketRequestHandler::send( const ByteBuffer& requestMsg )
{
retry:
  try {
    boost::mutex::scoped_lock lock( m_mutexForSend );
    m_requestID++;
    ReplySlot& slot=m_rplMngr->requestSlot( m_requestID );

    PacketSLAC packet(  m_requestID, 0 /*segment number*/, 0 /*clientID*/,
        false /* more segments*/, requestMsg );

    DEBUG("sending request #"<< m_requestID);
    m_Socket->sendPacket( packet );
    return slot;
  }
  catch ( AllSlotsTakenException &e ) {
    ERROR("(non fatal) all slots for requests are taken. Sleeping.");
    sleep( 20 );
    goto retry;
  }
  catch( std::exception& e )
  {
    ERROR("ERROR! Standard C++ exception: '" << e.what() << "'");
    throw;
  }
  catch( ... ) 
  {
    ERROR("ERROR! Unknown exception caught");
    throw;
 }
}

//-----------------------------------------------------------------------------

