##############################################################
# runHLT_standalone.py
#
#   jobOption file to run HLT standalone (independant of RecExCommon)
#   allows to run both cosmic and physics menus 
#
# Input could be Pool or BS
#
# Input file: (athena only)
#   Pool: set by "PoolRDOInput"
#   BS:   set by "BSRDOInput"
#
# L1 configuration is fixed
#   set by "setLVL1XML" 
#
# HLT configuration uses Python configuration
#   set by "setMenu" and "setPrescale"
#   Individual slices can be run by doing "test<sliceName>=True"
#   Individual signatures can be removed/added putting them in
#     "removeSignatures" and "addSignatures"
#     e.g. "addSignatures={'MuonSlice': ['mu6_ef']}"
#
# Geometry and conditions
#   CommisDB   set by "useCOMCONDDB"
#   Geometry   set by "setDetDescr"
#   Conditions set by "setGlobalTag"
#   Field      set by "setField" (possible values: 'All','Toroids','Solenoid','None')
#
# Special modifiers
#   Currently certain options needs to be modified depending on the type of data
#   Good defaults are provided for running on physics MC and cosmics Data
#   Change by specifying list in "setModifiers" or turn modifiers on/off
#    by doing "myModifier=True/False"
#   Existing modifiers can be found in "TriggerRelease/python/Modifiers.py"
#
# Predefined setups
#   certain menu configurations can be run just by doing
#    testSingleBeamV1=True or
#    testCollisionV1MC=True
#    testCosmicV1=True
#    test1E31MC=True
#   Note that for the addition of "MC" option. This is used for different setups.
#   Older options of not setting a menu or running the special cosmic setup testCosmic=True
#    is depreciated, but should still work.
#
# Usage:
#   run with BS input (default): athena.py -c test1E31MC=True runHLT_standalone.py >& testReadBS.log  &
#   run with Pool input: athena.py -c "ReadBS=False;test1E31MC=True" runHLT_standalone.py >& testTransBS.log  &
#   run LVL2 only: athena.py -s -c "doEF=False;test1E31MC=True" runHLT_standalone.py >& testReadBS_LVL2.log  &
#  
#   run with athenaMT:
#      ln -sf /afs/cern.ch/atlas/project/trigger/pesa-sw/releases/data/daq.physics1E31.May12devval.105200.Athena._0001.data input.data
#      athenaMT -f input.data -c test1E31MC=True -l DEBUG -n 25 runHLT_standalone.py >& testMT.log  &
#      or with multiple files
#      athenaMT -f "['input1.data','input2.data']" -c test1E31MC=True -n 25 runHLT_standalone.py >& testMT.log  &
#      or, for writing BS output:
#      athenaMT -f input.data -c test1E31MC=True -n 25 -o outBS runHLT_standalone.py >& testMT.log &
#      or, with online THistSvc
#      athenaMT -M -l DEBUG -c test1E31MC=True -n 25 runHLT_standalone.py >& testMT.log  &
#
#   run with athenaPT:
#      athenaPT -f outBS._0001.data -c test1E31MC=True -l DEBUG -n 25 runHLT_standalone.py >& testPT.log  &
#
#   run with EF only with athena:
#      ln -sf outBS_0001.data doEFOnly.input.data
#      athena.py -s -c "doLVL2=False;test1E31MC=True" runHLT_standalone.py >& testReadBS_EF.log  &
#
# Select slice(s) to test:
#   set one or more of the following flags to True in the jo:
#      testEgamma, testMuon, testTau, testJet, testBjet, testMET, testBphysics
#    e.g.:
#      athenaMT -f input.data -l DEBUG -n 25 -c "testTau=True;test1E31MC=True" runHLT_standalone.py >& testTauMT.log  &
#
#
# Contact: alessandro.cerri@cern,ch, Brian.Petersen@cern.ch
#===========================================================================================
from AthenaCommon.Logging import logging
log = logging.getLogger('runHLT_standalone.py')

if not ('testCosmic' in dir()):
    testCosmic=False
else:
    log.fatal('testCosmic is no longer supported - please use testCosmicV1 instead')
    import sys
    sys.exit(1)
    
if not ('emptyMenu' in dir()):
    emptyMenu=False

if not ('setField' in dir()):  #we have the default here, since geometry tags depend on it
    setField='All'

#hack to support bad original choice of commandline option name:
#if nothing is specified, we use old 10^31 or testCosmics setup
menuMap={'SingleBeamV1': 'SingleBeam_v1',
         'CollisionV1':  'Collision_v1',
         'CosmicV1':     'Cosmic_v1',
         'Cosmic2009V1': 'Cosmic2009_v1',
         '1E31':         'Physics_lumi1E31',
         '1E32':         'lumi1E32',
         '1E33':         'lumi1E33'}

newMenuSetup=0
setupForMC=True
for name in menuMap:
    for key in ['','MC']:
        if ('test'+name+key in dir()):
            newMenuSetup+=1
            setMenu=menuMap[name]
            setupForMC=(key=='MC')
if newMenuSetup>1 or (newMenuSetup==1 and testCosmic==True):
    log.fatal('More than one menu requested')
if newMenuSetup==0 and testCosmic==False:
    log.fatal('No specific menu set requested - Please migrate to use "test1E31MC=True" and new BS file!')
    import sys
    sys.exit(1)

#setup options for standalone configurations

defaultOptions={  #defaults are set for old physics menu - gets overwritten below. To be removed
    'setMenu'          : setMenu,
    'setPrescale'      : '100',
    'setLVL1XML'       : 'TriggerMenuXML',
    'removeSignatures' : {},
    'addSignatures'    : {},
    'PoolRDOInput'     : ['/afs/cern.ch/atlas/project/trigger/pesa-sw/releases/data/mc08.105200.T1_McAtNlo_Jimmy.digit.RDO.e357_s462_d126_25evts.pool.root'],
    'BSRDOInput'       : ['/afs/cern.ch/atlas/project/trigger/pesa-sw/releases/data/daq.physics1E31.May12devval.105200.Athena._0001.data'],
    'OutputLevel'      : DEBUG,
    'HLTOutputLevel'   : DEBUG,
    'ReadBS'           : True,
    'EvtMax'           : 25,
    'SkipEvents'       :  0,
    }

if not setMenu.count('lumi'): #should try to get default xml files for startup menus
    if setMenu=='Cosmic2009_v1':
        defaultOptions['setLVL1XML']='TriggerMenuXML/LVL1config_Cosmic2009_v1_7-bit_trigger_types.xml'
    else:
        defaultOptions['setLVL1XML']='TriggerMenuXML/LVL1config_SingleBeam_v1_7-bit_trigger_types.xml'
    defaultOptions['setPrescale']=''

#set geometry and conditions
if not setupForMC:
    defaultOptions['useCOMCONDDB']=True
    defaultOptions['setDetDescr']='ATLAS-GEO-03-00-00'
    if setField=='Toroids':
        defaultOptions['setDetDescr']='ATLAS-GEO-04-00-00'
    elif setField=='Solenoid':
        defaultOptions['setDetDescr']='ATLAS-GEONTF-05-00-00'
    elif setField=='None':
        defaultOptions['setDetDescr']='ATLAS-GEONF-04-00-00'
    defaultOptions['setGlobalTag']='COMCOND-HLTC-001-00'
else:
    defaultOptions['useCOMCONDDB']=False
    defaultOptions['setDetDescr']='ATLAS-GEO-02-01-00'   
    defaultOptions['setGlobalTag']='OFLCOND-SIM-00-00-03'

#Modifiers is a bit of a mess to setup     
defaultOptions['setModifiers']=[#Common modifiers:
                                'TRTCalibFromDB',
                                'noLArCalibFolders',
                                'noDetailedTiming',
                                'algorithmStack',]
if setMenu.count('lumi'):
    defaultOptions['setModifiers']+=[
                                'RPCcablingHack'] #needed for startup as well?
else:
    defaultOptions['setModifiers']+=[
                                'noCSCReadout', #turn off CSC in both data and MC
                                'enableHotIDMasking',
                                'disableCaloAllSamples',
                                'softTRTsettings',
                                'openThresholdRPCCabling',
                              #special streaming setup
                                'enable7BitL1TTStreaming']
if setupForMC:
    defaultOptions['setModifiers']+=[
                                'ForceNewTRTDigiversion',
                                'TGCCablingFixOff',
                                'Force8FoldTGC',
                                'ForceMCSCTMapping']
else:
    defaultOptions['setModifiers']+=[
                              #for running with real data
                                #'ForceNewDataPixelMapping', #obsolete, should be default now
                                #'ForceDataSCTMappingSept08', #running from file configuration is broken
                                'ForceMuonDataType',
                                'CaloCellMasking',
                                #'removeLArDBFolders', #obsolete
                                #'removeSCTDBFolders',
                                'MdtCalibFromDB',
                              #For running with cosmic (and current) data:
                                'RPCCosmicData',
                                'BFieldReconfigure',
                                'TGCCablingFixOn',
                              #for tests with data:
                                #'ignoreErrorStream', #nothing goes to debug stream
                                #'inclusiveErrorStream', #errors go to both debug and physics streams
                                #'ignoreL1Vetos',  #also run L2 prescaled and disabled L1 items
                                #'disablePixels',
                                #'disableSCTBarrel',
                                ]


#we don't want anything further on to depend on these flags
del newMenuSetup
del setupForMC
#-------------------------------------------------------------
# Transfer flags into TriggerFlags
#-------------------------------------------------------------
from TriggerJobOpts.TriggerFlags import TriggerFlags

# To turn off Lvl2/EF for athena running - do not set these by hand when using athenaMT/PT
if 'doLVL2' in dir():
    TriggerFlags.doLVL2 = bool(doLVL2)
    del doLVL2
if 'doEF' in dir():
    TriggerFlags.doEF = bool(doEF)
    del doEF
# To extract the Trigger configuration
if "doDBConfig" in dir():
    TriggerFlags.Online.doDBConfig = bool(doDBConfig)
    del doDBConfig
if "trigBase" in dir():
    TriggerFlags.Online.doDBConfigBaseName = trigBase
    del trigBase

#-------------------------------------------------------------
# Setup options
#-------------------------------------------------------------
log.info('Setup options:')
if not TriggerFlags.doLVL2():
    defaultOptions['BSRDOInput']=["doEFOnly.input.data"]
for option in defaultOptions:
    if option in dir():
        print ' %20s = %s' % (option,str(globals()[option]))
    else:
        globals()[option]=defaultOptions[option]
        print ' %20s = (Default) %s' % (option,str(globals()[option]))

#-------------------------------------------------------------
# Modifiers
#-------------------------------------------------------------

modifierList=[]
import TriggerRelease.Modifiers
for mod in dir(TriggerRelease.Modifiers):
    if not hasattr(getattr(TriggerRelease.Modifiers,mod),'preSetup'): continue
    if mod in dir():  #allow turning on and off modifiers by variable of same name
        if globals()[mod]:
            setModifiers+=[mod]
        else:
            if mod in setModifiers: setModifiers.remove(mod)
    if mod in setModifiers:
        modifierList+=[getattr(TriggerRelease.Modifiers,mod)()]
        setModifiers.remove(mod)
if setModifiers:
    log.error('Unknown modifier(s): '+str(setModifiers))


#-------------------------------------------------------------
# Menu
#  - allow to run individual slices by simply doing testSliceName=True
#  - for more detailed control use "addSignatures" and "removeSignatures" directly
#-------------------------------------------------------------
sliceList=[]
slicesToRun=[]
for prop in dir(TriggerFlags):
    if prop[-5:]=='Slice':
        sliceName=prop
        sliceList+=[sliceName]
        if dir().count('test'+sliceName[:-5]) and globals()['test'+sliceName[:-5]]:
            slicesToRun+=[sliceName]

if slicesToRun or emptyMenu:
    slicesToDisable=set(sliceList).difference(slicesToRun)
    #make a list of slices to disable once the menu has been setup
    for slice in slicesToDisable:
        removeSignatures[slice]='ALL'
else:
    slicesToRun=['All']

# should maybe implement the actual filter with regular expressions for added flexibility 
def enableDisableChains():
    # this functions is called in menu generations after the
    # default signatures have been setup according to choice of menu
    log = logging.getLogger('runHLT_standalone.py')
    for s in sliceList:
        slice=getattr(TriggerFlags,s)
        if removeSignatures.has_key(s):
            remove=removeSignatures[s]
            log.warn('Disabling '+str(remove)+' of '+s)
            if remove=='ALL':
                if slice.signatures.get_Value():
                    slice.disableSignatures(slice.signatures.get_Value())
            else:
                slice.disableSignatures(remove)
        if addSignatures.has_key(s):
            add=addSignatures[s]
            slice.enableSignatures(add)
            log.info('Enabled '+str(add)+' in '+s)

from TriggerMenuPython.GenerateMenu import GenerateMenu
GenerateMenu.overwriteSignaturesWith(enableDisableChains)

if emptyMenu and not addSignatures:
    log.warn('Completely empty menu - adding single empty chain')
    def insertEmptyChain(config):
        from TrigConfigSvc.TriggerPythonConfig import HLTChain
        empty = HLTChain("AllAcceptingChain", "0", "", "L2")
        empty.addStreamTag("nohlt")
        config.addHLTChain(empty)
        pass
    GenerateMenu.overwriteMenuWith(insertEmptyChain)
    
log.info("=============== Running slices: ===============")
for s in slicesToRun:
    log.info(s)
log.info("===============================================")

#-------------------------------------------------------------
# Setting TriggerFlags
#-------------------------------------------------------------
include.block("RecExCond/RecExCommon_flags.py")

TriggerFlags.doHLTpersistency=False
TriggerFlags.writeBS=True
TriggerFlags.abortOnConfigurationError=True

TriggerFlags.triggerMenuSetup=setMenu
if setPrescale:
    TriggerFlags.L1PrescaleSet  = 'L1Prescales' +setPrescale+'_'+setMenu
    TriggerFlags.HLTPrescaleSet = 'HLTPrescales'+setPrescale+'_'+setMenu
else:    
    TriggerFlags.L1PrescaleSet  = 'None'
    TriggerFlags.HLTPrescaleSet = 'None'
    
# LVL1 config
TriggerFlags.readLVL1configFromXML=True
if setLVL1XML=='TriggerMenuXML':
    setLVL1XML='TriggerMenuXML/LVL1config_'+setMenu+'_15.2.0.xml' #pick right L1 menu
TriggerFlags.inputLVL1configFile=setLVL1XML
TriggerFlags.outputLVL1configFile=None

# HLT config
TriggerFlags.enableMonitoring = ['Online','Time']
TriggerFlags.readHLTconfigFromXML=False

# define a unique HLT configuration xml when running in a partition
import os
if os.environ.has_key("TDAQ_PARTITION"):
    from TriggerRelease import Utils
    ident = Utils.uniqueFileName()
    TriggerFlags.outputHLTconfigFile ="outputHLTconfig_%s.xml" % ident
    TriggerFlags.inputHLTconfigFile  ="outputHLTconfig_%s.xml" % ident

### detector flags - except for cosmics we now turn all detectors on by default
TriggerFlags.doID=True
TriggerFlags.doMuon=True
TriggerFlags.doCalo=True
from AthenaCommon.AthenaCommonFlags import athenaCommonFlags
TriggerFlags.MuonSlice.doMuonCalibrationStream = athenaCommonFlags.isOnline()

TriggerFlags.CosmicSlice.testCosmic=testCosmic #Makes cosmic slice quiet when not used

log.info("Trigger xml files: %s %s %s" % (TriggerFlags.outputHLTconfigFile,TriggerFlags.outputLVL1configFile,TriggerFlags.inputHLTconfigFile))

#-------------------------------------------------------------
# Setting Global Flags
#-------------------------------------------------------------
from AthenaCommon.GlobalFlags import globalflags

if useCOMCONDDB:
    globalflags.DetGeo='commis'
    globalflags.DataSource='data'
    if setGlobalTag.find('COM')==-1:
        log.error('Using commissioning DB, but GlobalTag='+setGlobalTag)
else:
    globalflags.DetGeo='atlas'
    globalflags.DataSource='geant4'
    if setDetDescr.find('Comm')!=-1:
        log.error('Using offline DB, but DetDesc='+setDetDescr)
    if setGlobalTag.find('COM')!=-1:
        log.error('Using offline DB, but GlobalTag='+setGlobalTag)
globalflags.DetDescrVersion=setDetDescr
globalflags.Luminosity='zero'
globalflags.ConditionsTag=setGlobalTag

#set new field config flags. Not sure how much they affect the trigger
from AthenaCommon.BFieldFlags import jobproperties
if setField=='Toroid':
    jobproperties.BField.solenoidOn=False
    jobproperties.BField.barrelToroidOn=True
    jobproperties.BField.endcapToroidOn=True
elif setField=='Solenoid':
    jobproperties.BField.solenoidOn=True
    jobproperties.BField.barrelToroidOn=False
    jobproperties.BField.endcapToroidOn=False
elif setField=='None':
    jobproperties.BField.solenoidOn=False
    jobproperties.BField.barrelToroidOn=False
    jobproperties.BField.endcapToroidOn=False
else:
    if setField!='All':
        log.error('Unknown field configuration: setField='+setField)
    jobproperties.BField.solenoidOn=True
    jobproperties.BField.barrelToroidOn=True
    jobproperties.BField.endcapToroidOn=True

# Input format and file
if not ReadBS:
    globalflags.InputFormat='pool'
    if type(PoolRDOInput)==type(''):
        athenaCommonFlags.PoolRDOInput=[PoolRDOInput]
    else:
        athenaCommonFlags.PoolRDOInput=PoolRDOInput
else :
    globalflags.InputFormat='bytestream'
    if type(BSRDOInput)==type(''):
        athenaCommonFlags.BSRDOInput=[BSRDOInput]
    else:
        athenaCommonFlags.BSRDOInput=BSRDOInput

#-------------------------------------------------------------
# Apply modifiers
#-------------------------------------------------------------
for mod in modifierList:
    mod.preSetup()

#-------------------------------------------------------------
# Setup trigger
#-------------------------------------------------------------
include("TriggerRelease/Trigger_topOptions_standalone.py")
    
#-------------------------------------------------------------
# Apply modifiers
#-------------------------------------------------------------
for mod in modifierList:
    mod.postSetup()

# ----------------------------------------------------------------
# Number of events to be processed - for athena
# ----------------------------------------------------------------
theApp.EvtMax = EvtMax
if hasattr(svcMgr,"EventSelector"):
    svcMgr.EventSelector.SkipEvents = SkipEvents 

print AlgSequence
print ServiceMgr

print "_______________________________"
print "A73X: Full JO read successfully"
print "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"

# See https://twiki.cern.ch/twiki/bin/view/Atlas/StartingDebuggerWithAthenaPy
theApp.getHandle().Dlls += ['IOVDbSvc'] # i.e. libIOVDbSvc.so
from AthenaCommon.Debugging import *
###help(hookDebugger)
hookDebugger()

# Interactive gdb session
###(gdb) sha IOVDbSvc
###(gdb) info signals
###(gdb) dir /afs/cern.ch/atlas/software/builds/AtlasCore/15.2.0/Database/IOVDbSvc/src
###(gdb) br 'IOVDbFolder::loadCache(unsigned long long, unsigned int, std::string const&, bool)' 
###(gdb) br IOVDbFolder.cxx:356
###(gdb) display dpPtr
###(gdb) c

###import time
###time.sleep(10)
