#include "ConnectionServiceConfiguration.h"
#include "RelationalAccess/IRelationalService.h"
#include "RelationalAccess/IAuthenticationService.h"
#include "RelationalAccess/ILookupService.h"
#include "RelationalAccess/IMonitoringService.h"
#include "RelationalAccess/ConnectionServiceException.h"
#include "CoralBase/MessageStream.h"
#include "CoralKernel/Context.h"

#include <stdexcept>
#include <sstream>

/// constructor
coral::ConnectionService::ConnectionServiceConfiguration::ConnectionServiceConfiguration( const std::string& serviceName ):
  IDevConnectionServiceConfiguration(),
  m_serviceContext( coral::Context::instance() ),
  m_serviceName( serviceName ),
  m_replicaFailover( true ),
  m_connectionSharing( true ),
  m_readOnlyOnUpdate( true ),
  m_automaticCleanUp( true ),
  m_periodOfConnectionRetrials( 10 ),
  m_connectionRetrialTimeOut( 60 ),
  m_connectionTimeOut( 300 ),
  m_exclusionTime( 3600 ),
  m_monitoringLevel( coral::monitor::Off ),
  m_authenticationService( "" ), 
  m_lookupService( "" ),
  m_monitoringService( "CORAL/Services/MonitoringService" ),
  m_relationalService( "CORAL/Services/RelationalService" ),
  m_customAuthenticationService(0),
  m_customLookupService(0),
  m_customRelationalService(0),
  m_customMonitoringService(0),
  m_sortingAlgorithm(0),
  m_poolCleanUpPeriod( 10 )
{

  coral::MessageStream log( m_serviceName );

  char* userAuthSvc = ::getenv ( "CORAL_AUTH_SERVICE_LIBRARY" );
  if ( userAuthSvc ) {
    log << coral::Debug << "CORAL_AUTH_SERVICE_LIBRARY sets user-defined authentication service '" << userAuthSvc << "'" << coral::MessageStream::endmsg;    
    m_authenticationService = "CORAL/Services/UserDefinedAuthenticationService";
  }
  else if ( ::getenv ( "CORAL_AUTH_PATH" ) ) {
    m_authenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  else {
    if ( ::getenv ( "CORAL_AUTH_USER" ) && ::getenv ( "CORAL_AUTH_PASSWORD" ) )
      m_authenticationService = "CORAL/Services/EnvironmentAuthenticationService";
    else
      m_authenticationService = "CORAL/Services/XMLAuthenticationService";
  }
  
  char* userLookupSvc = ::getenv ( "CORAL_LOOKUP_SERVICE_LIBRARY" );
  if ( userLookupSvc ) {
    log << coral::Debug << "CORAL_LOOKUP_SERVICE_LIBRARY sets user-defined lookup service '" << userLookupSvc << "'" << coral::MessageStream::endmsg;    
    m_lookupService = "CORAL/Services/UserDefinedLookupService";
  }
  else {
    m_lookupService = "CORAL/Services/XMLLookupService";
  }

  if ( m_connectionSharing )
    log << coral::Debug << "Connection sharing is enabled" << coral::MessageStream::endmsg;
  else
    log << coral::Debug << "Connection sharing is disabled" << coral::MessageStream::endmsg;
}

/// destructor
coral::ConnectionService::ConnectionServiceConfiguration::~ConnectionServiceConfiguration(){
  // The plug-ins constructed by the users are NOT owned by CORAL.
}

/// enables the sharing of the same physical connection among more clients.
void 
coral::ConnectionService::ConnectionServiceConfiguration::enableConnectionSharing(){
  m_connectionSharing = true;
  coral::MessageStream log( m_serviceName );
  log << coral::Debug << "Enable connection sharing" << coral::MessageStream::endmsg;
}

/// disables the sharing of the same physical connection among more clients.
void 
coral::ConnectionService::ConnectionServiceConfiguration::disableConnectionSharing(){
  m_connectionSharing = false;
  coral::MessageStream log( m_serviceName );
  log << coral::Debug << "Disable connection sharing" << coral::MessageStream::endmsg;
}

/// loads the required plugins according to defaults
coral::IRelationalService&
coral::ConnectionService::ConnectionServiceConfiguration::relationalService() const 
{
  if(m_customRelationalService){
    return *m_customRelationalService;
  }

  coral::IHandle<coral::IRelationalService> relSvc = coral::Context::instance().query<coral::IRelationalService>( m_relationalService );
  
  if ( ! relSvc.isValid() )
  {  
    coral::MessageStream log( m_serviceName );
    log << coral::Info << "Loading default plugin for coral::IRelationalService: " <<m_relationalService << coral::MessageStream::endmsg;
    loadService( m_relationalService );
    relSvc = m_serviceContext.query<coral::IRelationalService>( m_relationalService );
    if( ! relSvc.isValid() )
    {
      std::ostringstream mess;
      mess << "Selected plugin "<< m_relationalService <<" has not been found.";
      log << coral::Error << mess.str() << coral::MessageStream::endmsg;
      
      throw ConnectionServiceException(mess.str(), "ConnectionService::ConnectionServiceConfiguration::relationalService");
    }                                                            
  }
  
  return *relSvc;
}

/// loads the required plugins according to defaults
coral::IAuthenticationService&
coral::ConnectionService::ConnectionServiceConfiguration::authenticationService() const
{
  if(m_customAuthenticationService){
    return *m_customAuthenticationService;
  }

  coral::IHandle<coral::IAuthenticationService> authSvc = m_serviceContext.query<coral::IAuthenticationService>( m_authenticationService );
  
  if ( ! authSvc.isValid() )
  {
    coral::MessageStream log( m_serviceName );
    log << coral::Info << "Loading default plugin for coral::IAuthenticationService: " << m_authenticationService << coral::MessageStream::endmsg;    
    loadService( m_authenticationService );
    authSvc = m_serviceContext.query<coral::IAuthenticationService>( m_authenticationService );
    
    if ( ! authSvc.isValid() )
    {
      std::ostringstream mess;
      mess << "Selected plugin "<< m_authenticationService <<" has been not found.";
      log << coral::Error << mess.str() << coral::MessageStream::endmsg;
      
      throw ConnectionServiceException(mess.str(), "ConnectionService::ConnectionServiceConfiguration::authenticationService");
    }                                                            
  }
  
  return *authSvc;
}

/// loads the required plugins according to defaults
coral::ILookupService&
coral::ConnectionService::ConnectionServiceConfiguration::lookupService() const
{
  if(m_customLookupService){
    return *m_customLookupService;
  }

  coral::IHandle<coral::ILookupService> lookupSvc = m_serviceContext.query<coral::ILookupService>( m_lookupService );
  
  if ( ! lookupSvc.isValid() )
  {
    coral::MessageStream log( m_serviceName );
    log << coral::Info << "Loading default plugin for coral::ILookupService: " << m_lookupService << coral::MessageStream::endmsg;    
    loadService( m_lookupService );
    lookupSvc = m_serviceContext.query<coral::ILookupService>( m_lookupService );
    if( ! lookupSvc.isValid() )
    {
      std::ostringstream mess;
      mess << "Selected plugin "<< m_lookupService <<" has been not found.";
      log << coral::Error << mess.str() << coral::MessageStream::endmsg;
      
      throw ConnectionServiceException(mess.str(), "ConnectionService::ConnectionServiceConfiguration::lookupService");
    }                                                            
  }
  
  return *lookupSvc;
}

/// loads the required plugin according to defaults
coral::monitor::IMonitoringService&
coral::ConnectionService::ConnectionServiceConfiguration::monitoringService() const
{
  if(m_customMonitoringService){
    return *m_customMonitoringService;
  }

  coral::IHandle<coral::monitor::IMonitoringService> monSvc = m_serviceContext.query<coral::monitor::IMonitoringService>( m_monitoringService );
  
  if ( ! monSvc.isValid() )
  {
    coral::MessageStream log( m_serviceName );
    log << coral::Info << "Loading default plugin for coral::IMonitoringService: " << m_monitoringService << coral::MessageStream::endmsg;
    loadService( m_monitoringService );
    monSvc = m_serviceContext.query<coral::monitor::IMonitoringService>( m_monitoringService );
    if( ! monSvc.isValid() )
    {
      std::ostringstream mess;
      mess << "Selected plugin "<< m_monitoringService <<" has been not found.";
      log << coral::Error << mess.str() << coral::MessageStream::endmsg;
      
      throw ConnectionServiceException(mess.str(), "ConnectionService::ConnectionServiceConfiguration::monitoringService");
    }                                                            
  }
  
  return *monSvc;
}

/**
 * Loads and sets the authentication service to be used for the new sessions.
 */
void
coral::ConnectionService::ConnectionServiceConfiguration::setAuthenticationService( const std::string serviceName) {
  m_authenticationService = serviceName;
  m_customAuthenticationService = 0;
  coral::MessageStream log( m_serviceName );
  log << coral::Info << "Loading default plugin for coral::IAuthenticationService: " << m_authenticationService << coral::MessageStream::endmsg;
  loadService( m_authenticationService );
}

/**
 * Loads and sets the lookup service to be used for the new sessions.
 */
void
coral::ConnectionService::ConnectionServiceConfiguration::setLookupService( const std::string& serviceName ){
  m_lookupService = serviceName;
  m_customLookupService = 0;
  coral::MessageStream log( m_serviceName );
  log << coral::Info << "Loading default plugin for coral::ILookupService: " << m_lookupService << coral::MessageStream::endmsg;
  loadService( m_lookupService );
}

/**
 * Loads and sets the relational service to be used for the new sessions.
 */
void
coral::ConnectionService::ConnectionServiceConfiguration::setRelationalService( const std::string& serviceName ){
  m_relationalService = serviceName;
  m_customRelationalService = 0;
  coral::MessageStream log( m_serviceName );
  log << coral::Info << "Loading default plugin for coral::IRelationalService: " << m_relationalService << coral::MessageStream::endmsg;
  loadService( m_relationalService );  
}     

/**
 * Loads and sets the monitoring service to be used for the new sessions.
 */
void
coral::ConnectionService::ConnectionServiceConfiguration::setMonitoringService( const std::string& serviceName ){
  m_monitoringService = serviceName;
  m_customMonitoringService = 0;
  coral::MessageStream log( m_serviceName );
  log << coral::Info << "Loading default plugin for coral::IMonitoringService: " << m_monitoringService << coral::MessageStream::endmsg;
  loadService( m_monitoringService );  
}

/// access to the service Context
coral::Context*
coral::ConnectionService::ConnectionServiceConfiguration::serviceContext() const
{
  return &m_serviceContext;
}

/// access to the upper levevl service name
const std::string&
coral::ConnectionService::ConnectionServiceConfiguration::serviceName() const
{
  return m_serviceName;
}

/// loads the the specified Service
void
coral::ConnectionService::ConnectionServiceConfiguration::loadService( const std::string& serviceName ) const
{
  // Load a CORAL service by name
  coral::Context::instance().loadComponent( serviceName );
}

